import os
import warnings
from functools import partial
from typing import Callable, Mapping, Optional, Type, Union

import numpy as np
import tensorflow as tf
import tqdm
from stable_baselines.common import base_class, vec_env

from imitation.data import buffer, datasets, types, wrappers
from imitation.rewards import discrim_net, reward_net
from imitation.util import logger, reward_wrapper


class AdversarialTrainer:
    """Base class for adversarial imitation learning algorithms like GAIL and AIRL."""

    venv: vec_env.VecEnv
    """The original vectorized environment."""

    venv_train: vec_env.VecEnv
    """Like `self.venv`, but wrapped with train reward unless in debug mode.

    If `debug_use_ground_truth=True` was passed into the initializer then
    `self.venv_train` is the same as `self.venv`."""

    venv_test: vec_env.VecEnv
    """Like `self.venv`, but wrapped with test reward unless in debug mode.

    If `debug_use_ground_truth=True` was passed into the initializer then
    `self.venv_test` is the same as `self.venv`."""

    def __init__(
        self,
        venv: vec_env.VecEnv,
        gen_policy: base_class.BaseRLModel,
        discrim: discrim_net.DiscrimNet,
        expert_data: Union[datasets.Dataset[types.Transitions], types.Transitions],
        *,
        log_dir: str = "output/",
        disc_batch_size: int = 2048,
        disc_minibatch_size: int = 256,
        disc_opt_cls: Type[tf.train.Optimizer] = tf.train.AdamOptimizer,
        disc_opt_kwargs: Optional[Mapping] = None,
        gen_replay_buffer_capacity: Optional[int] = None,
        init_tensorboard: bool = False,
        init_tensorboard_graph: bool = False,
        debug_use_ground_truth: bool = False,
    ):
        """Builds AdversarialTrainer.

        Args:
            venv: The vectorized environment to train in.
            gen_policy: The generator policy that is trained to maximize
              discriminator confusion. The generator batch size
              `self.gen_batch_size` is inferred from `gen_policy.n_batch`.
            discrim: The discriminator network.
            expert_data: Either a `Dataset` of expert `Transitions`, or an instance of
                `Transitions` to be automatically converted into a
                `Dataset[Transitions]`.
            log_dir: Directory to store TensorBoard logs, plots, etc. in.
            disc_batch_size: The default number of expert and generator transitions
              samples to feed to the discriminator in each call to
              `self.train_disc()`. (Half of the samples are expert and half of the
              samples are generator).
            disc_minibatch_size: The discriminator minibatch size. Each
              discriminator batch is split into minibatches and an Adam update is
              applied on the gradient resulting form each minibatch. Must evenly
              divide `disc_batch_size`. Must be an even number.
            disc_opt_cls: The optimizer for discriminator training.
            disc_opt_kwargs: Parameters for discriminator training.
            gen_replay_buffer_capacity: The capacity of the
              generator replay buffer (the number of obs-action-obs samples from
              the generator that can be stored).

              By default this is equal to `self.gen_batch_size`, meaning that we
              sample only from the most recent batch of generator samples.
            init_tensorboard: If True, makes various discriminator
              TensorBoard summaries.
            init_tensorboard_graph: If both this and `init_tensorboard` are True,
              then write a Tensorboard graph summary to disk.
            debug_use_ground_truth: If True, use the ground truth reward for
              `self.train_env`.
              This disables the reward wrapping that would normally replace
              the environment reward with the learned reward. This is useful for
              sanity checking that the policy training is functional.
        """
        assert (
            logger.is_configured()
        ), "Requires call to imitation.util.logger.configure"
        self._sess = tf.get_default_session()
        self._global_step = tf.train.create_global_step()

        assert disc_batch_size % disc_minibatch_size == 0
        assert disc_minibatch_size % 2 == 0, (
            "discriminator minibatch size must be even "
            "(equal split between generator and expert samples)"
        )
        self.disc_batch_size = disc_batch_size
        self.disc_minibatch_size = disc_minibatch_size
        self.debug_use_ground_truth = debug_use_ground_truth
        self.venv = venv
        self._gen_policy = gen_policy
        self._log_dir = log_dir

        # Create graph for optimising/recording stats on discriminator
        self._discrim = discrim
        self._disc_opt_cls = disc_opt_cls
        self._disc_opt_kwargs = disc_opt_kwargs or {}
        self._init_tensorboard = init_tensorboard
        self._init_tensorboard_graph = init_tensorboard_graph
        self._build_graph()
        self._sess.run(tf.global_variables_initializer())

        if debug_use_ground_truth:
            # Would use an identity reward fn here, but RewardFns can't see rewards.
            self.reward_train = self.reward_test = None
            self.venv_train = self.venv_test = self.venv
        else:
            self.reward_train = partial(
                self.discrim.reward_train,
                # The generator policy uses normalized observations
                # but the reward function (self.reward_train) and discriminator use
                # and receive unnormalized observations. Therefore to get the right
                # log action probs for AIRL's ent bonus, we need to normalize obs.
                gen_log_prob_fn=self._gen_log_action_prob_from_unnormalized,
            )
            self.reward_test = self.discrim.reward_test
            self.venv_train = reward_wrapper.RewardVecEnvWrapper(
                self.venv, self.reward_train
            )
            self.venv_test = reward_wrapper.RewardVecEnvWrapper(
                self.venv, self.reward_test
            )

        self.venv_train_buffering = wrappers.BufferingWrapper(self.venv_train)
        self.venv_train_norm = vec_env.VecNormalize(self.venv_train_buffering)
        self.gen_policy.set_env(self.venv_train_norm)

        if gen_replay_buffer_capacity is None:
            gen_replay_buffer_capacity = self.gen_batch_size
        self._gen_replay_buffer = buffer.ReplayBuffer(
            gen_replay_buffer_capacity, self.venv
        )

        if isinstance(expert_data, types.Transitions):
            # Somehow, pytype doesn't recognize that `expert_data` is Transitions.
            expert_data = datasets.TransitionsDictDatasetAdaptor(
                expert_data,  # pytype: disable=wrong-arg-types
            )
        self._expert_dataset = expert_data

        expert_ds_size = self.expert_dataset.size()
        if expert_ds_size is not None and self.disc_batch_size // 2 > expert_ds_size:
            warnings.warn(
                "The discriminator batch size is more than twice the number of "
                "expert samples. This means that we will be reusing expert samples "
                "every discrim batch.",
                category=RuntimeWarning,
            )

    @property
    def gen_batch_size(self) -> int:
        return self.gen_policy.n_batch

    @property
    def discrim(self) -> discrim_net.DiscrimNet:
        """Discriminator being trained, used to compute reward for policy."""
        return self._discrim

    @property
    def expert_dataset(self) -> datasets.Dataset[types.Transitions]:
        """Dataset containing expert demonstrations that are being imitated."""
        return self._expert_dataset

    @property
    def gen_policy(self) -> base_class.BaseRLModel:
        """Policy (i.e. the generator) being trained."""
        return self._gen_policy

    def _gen_log_action_prob_from_unnormalized(
        self,
        observation: np.ndarray,
        *,
        actions: np.ndarray,
        logp=True,
    ) -> np.ndarray:
        """Calculate generator log action probabilility.

        Params:
          observation: Unnormalized observation.
          actions: action.
        """
        obs = self.venv_train_norm.normalize_obs(observation)
        return self.gen_policy.action_probability(obs, actions=actions, logp=logp)

    def train_disc(self, n_samples: Optional[int] = None) -> None:
        """Trains the discriminator to minimize classification cross-entropy.

        Must call `train_gen` first (otherwise there will be no saved generator
        samples for training, and will error).

        Args:
          n_samples: A number of transitions to sample from the generator
            replay buffer and the expert demonstration dataset. (Half of the
            samples are from each source). By default, `self.disc_batch_size`.
            `n_samples` must be a positive multiple of `self.disc_minibatch_size`.
        """
        if self._gen_replay_buffer.size() == 0:
            raise RuntimeError(
                "No generator samples for training. " "Call `train_gen()` first."
            )

        if n_samples is None:
            n_samples = self.disc_batch_size
        n_updates = n_samples // self.disc_minibatch_size
        assert n_samples % self.disc_minibatch_size == 0
        assert n_updates >= 1
        for _ in range(n_updates):
            gen_samples = self._gen_replay_buffer.sample(self.disc_minibatch_size)
            self.train_disc_step(gen_samples=gen_samples)

    def train_disc_step(
        self,
        *,
        gen_samples: Optional[types.Transitions] = None,
        expert_samples: Optional[types.Transitions] = None,
    ) -> None:
        """Perform a single discriminator update, optionally using provided samples.

        Args:
          gen_samples: Transition samples from the generator policy. If not
            provided, then take `self.disc_batch_size // 2` samples from the
            generator replay buffer. Observations should not be normalized.
          expert_samples: Transition samples from the expert. If not
            provided, then take `n_gen` expert samples from the expert
            dataset, where `n_gen` is the number of samples in `gen_samples`.
            Observations should not be normalized.
        """
        with logger.accumulate_means("disc"):
            fetches = {
                "train_op_out": self._disc_train_op,
                "train_stats": self._discrim.train_stats,
            }
            # optionally write TB summaries for collected ops
            step = self._sess.run(self._global_step)
            write_summaries = self._init_tensorboard and step % 20 == 0
            if write_summaries:
                fetches["events"] = self._summary_op

            # do actual update
            fd = self._build_disc_feed_dict(
                gen_samples=gen_samples, expert_samples=expert_samples
            )
            fetched = self._sess.run(fetches, feed_dict=fd)

            if write_summaries:
                self._summary_writer.add_summary(fetched["events"], fetched["step"])

            logger.logkv("step", step)
            for k, v in fetched["train_stats"].items():
                logger.logkv(k, v)
            logger.dumpkvs()

    def eval_disc_loss(self, **kwargs) -> float:
        """Evaluates the discriminator loss.

        Args:
          gen_samples (Optional[rollout.Transitions]): Same as in `train_disc_step`.
          expert_samples (Optional[rollout.Transitions]): Same as in
            `train_disc_step`.

        Returns:
          The total cross-entropy error in the discriminator's classification.
        """
        fd = self._build_disc_feed_dict(**kwargs)
        return np.mean(self._sess.run(self.discrim.disc_loss, feed_dict=fd))

    def train_gen(
        self,
        total_timesteps: Optional[int] = None,
        learn_kwargs: Optional[Mapping] = None,
    ):
        """Trains the generator to maximize the discriminator loss.

        After the end of training populates the generator replay buffer (used in
        discriminator training) with `self.disc_batch_size` transitions.

        Args:
          total_timesteps: The number of transitions to sample from
            `self.venv_train_norm` during training. By default,
            `self.gen_batch_size`.
          learn_kwargs: kwargs for the Stable Baselines `RLModel.learn()`
            method.
        """
        if total_timesteps is None:
            total_timesteps = self.gen_batch_size
        if learn_kwargs is None:
            learn_kwargs = {}

        with logger.accumulate_means("gen"):
            self.gen_policy.learn(
                total_timesteps=total_timesteps,
                reset_num_timesteps=False,
                **learn_kwargs,
            )

        gen_samples = self.venv_train_norm.pop_transitions()
        self._gen_replay_buffer.store(gen_samples)

    def train(
        self,
        total_timesteps: int,
        callback: Optional[Callable[[int], None]] = None,
    ) -> None:
        """Alternates between training the generator and discriminator.

        Every epoch consists of a call to `train_gen(self.gen_batch_size)`,
        a call to `train_disc(self.disc_batch_size)`, and
        finally a call to `callback(epoch)`.

        Training ends once an additional epoch would cause the number of transitions
        sampled from the environment to exceed `total_timesteps`.

        Params:
          total_timesteps: An upper bound on the number of transitions to sample
            from the environment during training.
          callback: A function called at the end of every epoch which takes in a
            single argument, the epoch number. Epoch numbers are in
            `range(total_timesteps // self.gen_batch_size)`.
        """
        n_epochs = total_timesteps // self.gen_batch_size
        assert n_epochs >= 1, (
            "No updates (need at least "
            f"{self.gen_batch_size} timesteps, have only "
            f"total_timesteps={total_timesteps})!"
        )
        for epoch in tqdm.tqdm(range(0, n_epochs), desc="epoch"):
            self.train_gen(self.gen_batch_size)
            self.train_disc(self.disc_batch_size)
            if callback:
                callback(epoch)
            logger.dumpkvs()

    def _build_graph(self):
        # Build necessary parts of the TF graph. Most of the real action happens in
        # constructors for self.discrim and self.gen_policy.
        with tf.variable_scope("trainer"):
            with tf.variable_scope("discriminator"):
                disc_opt = self._disc_opt_cls(**self._disc_opt_kwargs)
                self._disc_train_op = disc_opt.minimize(
                    tf.reduce_mean(self.discrim.disc_loss),
                    global_step=self._global_step,
                )

        if self._init_tensorboard:
            with tf.name_scope("summaries"):
                tf.logging.info("building summary directory at " + self._log_dir)
                graph = self._sess.graph if self._init_tensorboard_graph else None
                summary_dir = os.path.join(self._log_dir, "summary")
                os.makedirs(summary_dir, exist_ok=True)
                self._summary_writer = tf.summary.FileWriter(summary_dir, graph=graph)
                self._summary_op = tf.summary.merge_all()

    def _build_disc_feed_dict(
        self,
        *,
        gen_samples: Optional[types.Transitions] = None,
        expert_samples: Optional[types.Transitions] = None,
    ) -> dict:
        """Build and return feed dict for the next discriminator training update.

        Args:
          gen_samples: Same as in `train_disc_step`.
          expert_samples: Same as in `train_disc_step`.
        """
        if gen_samples is None:
            if self._gen_replay_buffer.size() == 0:
                raise RuntimeError(
                    "No generator samples for training. " "Call `train_gen()` first."
                )
            gen_samples = self._gen_replay_buffer.sample(self.disc_batch_size // 2)
        n_gen = len(gen_samples.obs)

        if expert_samples is None:
            expert_samples = self._expert_dataset.sample(n_gen)
        n_expert = len(expert_samples.obs)

        # Check dimensions.
        n_samples = n_expert + n_gen
        assert n_expert == len(expert_samples.acts)
        assert n_expert == len(expert_samples.next_obs)
        assert n_gen == len(gen_samples.acts)
        assert n_gen == len(gen_samples.next_obs)

        # Policy and reward network were trained on normalized observations.
        expert_obs_norm = self.venv_train_norm.normalize_obs(expert_samples.obs)
        gen_obs_norm = self.venv_train_norm.normalize_obs(gen_samples.obs)

        # Concatenate rollouts, and label each row as expert or generator.
        obs = np.concatenate([expert_obs_norm, gen_obs_norm])
        acts = np.concatenate([expert_samples.acts, gen_samples.acts])
        next_obs = np.concatenate([expert_samples.next_obs, gen_samples.next_obs])
        dones = np.concatenate([expert_samples.dones, gen_samples.dones])
        labels_gen_is_one = np.concatenate(
            [np.zeros(n_expert, dtype=int), np.ones(n_gen, dtype=int)]
        )

        # Calculate generator-policy log probabilities.
        log_act_prob = self._gen_policy.action_probability(obs, actions=acts, logp=True)
        assert len(log_act_prob) == n_samples
        log_act_prob = log_act_prob.reshape((n_samples,))

        fd = {
            self.discrim.obs_ph: obs,
            self.discrim.act_ph: acts,
            self.discrim.next_obs_ph: next_obs,
            self.discrim.done_ph: dones,
            self.discrim.labels_gen_is_one_ph: labels_gen_is_one,
            self.discrim.log_policy_act_prob_ph: log_act_prob,
        }
        return fd


class GAIL(AdversarialTrainer):
    def __init__(
        self,
        venv: vec_env.VecEnv,
        expert_data: Union[types.Transitions, datasets.Dataset[types.Transitions]],
        gen_policy: base_class.BaseRLModel,
        *,
        discrim_kwargs: Optional[Mapping] = None,
        **kwargs,
    ):
        """Generative Adversarial Imitation Learning.

        Most parameters are described in and passed to `AdversarialTrainer.__init__`.
        Additional parameters that `GAIL` adds on top of its superclass initializer are
        as follows:

        Args:
            discrim_kwargs: Optional keyword arguments to use while constructing the
                DiscrimNetGAIL.

        """
        discrim_kwargs = discrim_kwargs or {}
        discrim = discrim_net.DiscrimNetGAIL(
            venv.observation_space, venv.action_space, **discrim_kwargs
        )
        super().__init__(venv, gen_policy, discrim, expert_data, **kwargs)


class AIRL(AdversarialTrainer):
    def __init__(
        self,
        venv: vec_env.VecEnv,
        expert_data: Union[types.Transitions, datasets.Dataset[types.Transitions]],
        gen_policy: base_class.BaseRLModel,
        *,
        reward_net_cls: Type[reward_net.RewardNet] = reward_net.BasicShapedRewardNet,
        reward_net_kwargs: Optional[Mapping] = None,
        discrim_kwargs: Optional[Mapping] = None,
        **kwargs,
    ):
        """Adversarial Inverse Reinforcement Learning.

        Most parameters are described in and passed to `AdversarialTrainer.__init__`.
        Additional parameters that `AIRL` adds on top of its superclass initializer are
        as follows:

        Args:
            reward_net_cls: Reward network constructor. The reward network is part of
                the AIRL discriminator.
            reward_net_kwargs: Optional keyword arguments to use while constructing
                the reward network.
            discrim_kwargs: Optional keyword arguments to use while constructing the
                DiscrimNetAIRL.
        """
        # TODO(shwang): Maybe offer str=>Type[RewardNet] conversion like
        #  stable_baselines does with policy classes.
        reward_net_kwargs = reward_net_kwargs or {}
        reward_network = reward_net_cls(
            action_space=venv.action_space,
            observation_space=venv.observation_space,
            # pytype is afraid that we'll directly call RewardNet() which is an abstract
            # class, hence the disable.
            **reward_net_kwargs,  # pytype: disable=not-instantiable
        )

        discrim_kwargs = discrim_kwargs or {}
        discrim = discrim_net.DiscrimNetAIRL(reward_network, **discrim_kwargs)
        super().__init__(venv, gen_policy, discrim, expert_data, **kwargs)



"""Helper methods for tests of custom Gym environments.

This is used in the imitation test suite and may also be useful for users of this
library.
"""

import gym


def test_model_based(env: gym.Env) -> None:
    """Smoke test for each of the ModelBasedEnv methods with type checks.

    Raises:
        AssertionError if test fails.
    """
    state = env.initial_state()
    assert env.state_space.contains(state)

    action = env.action_space.sample()
    new_state = env.transition(state, action)
    assert env.state_space.contains(new_state)

    reward = env.reward(state, action, new_state)
    assert isinstance(reward, float)

    done = env.terminal(state, 0)
    assert isinstance(done, bool)

    obs = env.obs_from_state(state)
    assert env.observation_space.contains(obs)
    next_obs = env.obs_from_state(new_state)
    assert env.observation_space.contains(next_obs)



import contextlib
import os
import os.path as osp
from typing import Optional

import stable_baselines.logger as sb_logger
import tensorflow as tf
from sacred.observers import FileStorageObserver
from stable_baselines.common.vec_env import VecNormalize

import imitation.util.sacred as sacred_util
from imitation.data import rollout
from imitation.policies import serialize
from imitation.rewards.serialize import load_reward
from imitation.scripts.config.expert_demos import expert_demos_ex
from imitation.util import logger, networks, util
from imitation.util.reward_wrapper import RewardVecEnvWrapper


@expert_demos_ex.main
def rollouts_and_policy(
    _run,
    _seed: int,
    env_name: str,
    total_timesteps: int,
    *,
    log_dir: str,
    num_vec: int,
    parallel: bool,
    max_episode_steps: Optional[int],
    normalize: bool,
    normalize_kwargs: dict,
    init_rl_kwargs: dict,
    n_episodes_eval: int,
    reward_type: Optional[str],
    reward_path: Optional[str],
    rollout_save_interval: int,
    rollout_save_final: bool,
    rollout_save_n_timesteps: Optional[int],
    rollout_save_n_episodes: Optional[int],
    policy_save_interval: int,
    policy_save_final: bool,
    init_tensorboard: bool,
) -> dict:
    """Trains an expert policy from scratch and saves the rollouts and policy.

    Checkpoints:
      At applicable training steps `step` (where step is either an integer or
      "final"):

        - Policies are saved to `{log_dir}/policies/{step}/`.
        - Rollouts are saved to `{log_dir}/rollouts/{step}.pkl`.

    Args:
        env_name: The gym.Env name. Loaded as VecEnv.
        total_timesteps: Number of training timesteps in `model.learn()`.
        log_dir: The root directory to save metrics and checkpoints to.
        num_vec: Number of environments in VecEnv.
        parallel: If True, then use DummyVecEnv. Otherwise use SubprocVecEnv.
        max_episode_steps: If not None, then environments are wrapped by
            TimeLimit so that they have at most `max_episode_steps` steps per
            episode.
        normalize: If True, then rescale observations and reward.
        normalize_kwargs: kwargs for `VecNormalize`.
        init_rl_kwargs: kwargs for `init_rl`.

        n_episodes_eval: The number of episodes to average over when calculating
            the average ground truth reward return of the final policy.

        reward_type: If provided, then load the serialized reward of this type,
            wrapping the environment in this reward. This is useful to test
            whether a reward model transfers. For more information, see
            `imitation.rewards.serialize.load_reward`.
        reward_path: A specifier, such as a path to a file on disk, used by
            reward_type to load the reward model. For more information, see
            `imitation.rewards.serialize.load_reward`.

        rollout_save_interval: The number of training updates in between
            intermediate rollout saves. If the argument is nonpositive, then
            don't save intermediate updates.
        rollout_save_final: If True, then save rollouts right after training is
            finished.
        rollout_save_n_timesteps: The minimum number of timesteps saved in every
            file. Could be more than `rollout_save_n_timesteps` because
            trajectories are saved by episode rather than by transition.
            Must set exactly one of `rollout_save_n_timesteps`
            and `rollout_save_n_episodes`.
        rollout_save_n_episodes: The number of episodes saved in every
            file. Must set exactly one of `rollout_save_n_timesteps` and
            `rollout_save_n_episodes`.

        policy_save_interval: The number of training updates between saves. Has
            the same semantics are `rollout_save_interval`.
        policy_save_final: If True, then save the policy right after training is
            finished.

        init_tensorboard: If True, then write tensorboard logs to {log_dir}/sb_tb
            and "output/summary/...".

    Returns:
      The return value of `rollout_stats()` using the final policy.
    """
    os.makedirs(log_dir, exist_ok=True)
    sacred_util.build_sacred_symlink(log_dir, _run)

    sample_until = rollout.make_sample_until(
        rollout_save_n_timesteps, rollout_save_n_episodes
    )
    eval_sample_until = rollout.min_episodes(n_episodes_eval)

    with networks.make_session():
        tf.logging.set_verbosity(tf.logging.INFO)
        logger.configure(
            folder=osp.join(log_dir, "rl"), format_strs=["tensorboard", "stdout"]
        )

        rollout_dir = osp.join(log_dir, "rollouts")
        policy_dir = osp.join(log_dir, "policies")
        os.makedirs(rollout_dir, exist_ok=True)
        os.makedirs(policy_dir, exist_ok=True)

        if init_tensorboard:
            sb_tensorboard_dir = osp.join(log_dir, "sb_tb")
            # Convert sacred's ReadOnlyDict to dict so we can modify on next line.
            init_rl_kwargs = dict(init_rl_kwargs)
            init_rl_kwargs["tensorboard_log"] = sb_tensorboard_dir

        venv = util.make_vec_env(
            env_name,
            num_vec,
            seed=_seed,
            parallel=parallel,
            log_dir=log_dir,
            max_episode_steps=max_episode_steps,
        )

        log_callbacks = []
        with contextlib.ExitStack() as stack:
            if reward_type is not None:
                reward_fn_ctx = load_reward(reward_type, reward_path, venv)
                reward_fn = stack.enter_context(reward_fn_ctx)
                venv = RewardVecEnvWrapper(venv, reward_fn)
                log_callbacks.append(venv.log_callback)
                tf.logging.info(
                    f"Wrapped env in reward {reward_type} from {reward_path}."
                )

            vec_normalize = None
            if normalize:
                venv = vec_normalize = VecNormalize(venv, **normalize_kwargs)

            policy = util.init_rl(venv, verbose=1, **init_rl_kwargs)

            # Make callback to save intermediate artifacts during training.
            step = 0

            def callback(locals_: dict, _) -> bool:
                nonlocal step
                step += 1
                policy = locals_["self"]

                # TODO(adam): make logging frequency configurable
                for callback in log_callbacks:
                    callback(sb_logger)

                if rollout_save_interval > 0 and step % rollout_save_interval == 0:
                    save_path = osp.join(rollout_dir, f"{step}.pkl")
                    rollout.rollout_and_save(save_path, policy, venv, sample_until)
                if policy_save_interval > 0 and step % policy_save_interval == 0:
                    output_dir = os.path.join(policy_dir, f"{step:05d}")
                    serialize.save_stable_model(output_dir, policy, vec_normalize)

            policy.learn(total_timesteps, callback=callback)

            # Save final artifacts after training is complete.
            if rollout_save_final:
                save_path = osp.join(rollout_dir, "final.pkl")
                rollout.rollout_and_save(save_path, policy, venv, sample_until)
            if policy_save_final:
                output_dir = os.path.join(policy_dir, "final")
                serialize.save_stable_model(output_dir, policy, vec_normalize)

            # Final evaluation of expert policy.
            trajs = rollout.generate_trajectories(policy, venv, eval_sample_until)
            stats = rollout.rollout_stats(trajs)

    return stats


@expert_demos_ex.command
@networks.make_session()
def rollouts_from_policy(
    _run,
    _seed: int,
    *,
    num_vec: int,
    rollout_save_n_timesteps: int,
    rollout_save_n_episodes: int,
    log_dir: str,
    policy_path: str,
    policy_type: str,
    env_name: str,
    parallel: bool,
    rollout_save_path: str,
    max_episode_steps: Optional[int],
) -> None:
    """Loads a saved policy and generates rollouts.

    Unlisted arguments are the same as in `rollouts_and_policy()`.

    Args:
        policy_type: Argument to `imitation.policies.serialize.load_policy`.
        policy_path: Argument to `imitation.policies.serialize.load_policy`.
        rollout_save_path: Rollout pickle is saved to this path.
    """
    os.makedirs(log_dir, exist_ok=True)
    sacred_util.build_sacred_symlink(log_dir, _run)

    sample_until = rollout.make_sample_until(
        rollout_save_n_timesteps, rollout_save_n_episodes
    )

    venv = util.make_vec_env(
        env_name,
        num_vec,
        seed=_seed,
        parallel=parallel,
        log_dir=log_dir,
        max_episode_steps=max_episode_steps,
    )

    with serialize.load_policy(policy_type, policy_path, venv) as policy:
        rollout.rollout_and_save(rollout_save_path, policy, venv, sample_until)


def main_console():
    observer = FileStorageObserver(osp.join("output", "sacred", "expert_demos"))
    expert_demos_ex.observers.append(observer)
    expert_demos_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



import sacred

analysis_ex = sacred.Experiment("analyze")


@analysis_ex.config
def config():
    source_dir = "output/sacred/train"  # Recursively search in this directory
    skip_failed_runs = True  # Skip analysis for logs that have FAILED status
    run_name = None  # Restrict analysis to sacred logs with a certain run name
    env_name = None  # Restrict analysis to sacred logs with a certain env name
    csv_output_path = None  # Write output CSV to this path
    verbose = True  # Set to True to print analysis to stdout



from setuptools import find_packages, setup

import src.imitation  # pytype: disable=import-error

TF_VERSION = ">=1.15.0,<2.0"

TESTS_REQUIRE = [
    "seals~=0.1.0",
    "black",
    # remove pin once https://github.com/nedbat/coveragepy/issues/881 fixed
    "coverage==4.5.4",
    "codecov",
    "codespell",
    "flake8",
    "flake8-blind-except",
    "flake8-builtins",
    "flake8-debugger",
    "flake8-isort",
    "pytest",
    "pytest-cov",
    "pytest-shard",
    "pytest-xdist",
    "pytype",
]
DOCS_REQUIRE = [
    "sphinx",
    "sphinxcontrib-napoleon",
]
PARALLEL_REQUIRE = ["ray[debug,tune]~=0.8.5"]


def get_readme() -> str:
    """Retrieve content from README."""
    with open("README.md", "r") as f:
        return f.read()


setup(
    name="imitation",
    version=src.imitation.__version__,
    description="Implementation of modern IRL and imitation learning algorithms.",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    author="Center for Human-Compatible AI and Google",
    python_requires=">=3.7.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"imitation": ["py.typed", "envs/examples/airl_envs/assets/*.xml"]},
    install_requires=[
        "awscli",
        "cloudpickle>=0.5.5",
        "gym[classic_control]",
        "matplotlib",
        "numpy>=1.15",
        "tqdm",
        "scikit-learn>=0.21.2",
        "stable-baselines~=2.10.1",
        "jax~=0.1.66",
        "jaxlib~=0.1.47",
        "sacred~=0.8.1",
    ],
    tests_require=TESTS_REQUIRE,
    extras_require={
        "gpu": [f"tensorflow-gpu{TF_VERSION}"],
        "cpu": [f"tensorflow{TF_VERSION}"],
        # recommended packages for development
        "dev": [
            "autopep8",
            "ntfy[slack]",
            "ipdb",
            "isort~=5.0",
            "jupyter",
            "pytype",
            "codespell",
            # for convenience
            *TESTS_REQUIRE,
            *DOCS_REQUIRE,
        ],
        "test": TESTS_REQUIRE,
        "docs": DOCS_REQUIRE,
        "parallel": PARALLEL_REQUIRE,
    },
    entry_points={
        "console_scripts": [
            ("imitation-expert-demos=imitation.scripts.expert_demos" ":main_console"),
            "imitation-train=imitation.scripts.train_adversarial:main_console",
        ],
    },
    url="https://github.com/HumanCompatibleAI/imitation",
    license="MIT",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import numpy as np
from gym import utils
from gym.envs.mujoco import mujoco_env

from imitation.envs.examples.airl_envs.dynamic_mjc.mjc_models import point_mass_maze


class PointMazeEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(
        self,
        direction=1,
        maze_length=0.6,
        sparse_reward=False,
        no_reward=False,
        include_vel=False,
        episode_length=100,
    ):
        utils.EzPickle.__init__(self)
        self.sparse_reward = sparse_reward
        self.no_reward = no_reward
        self.include_vel = include_vel
        self.max_episode_length = episode_length
        self.direction = direction
        self.length = maze_length

        self.episode_length = 0

        model = point_mass_maze(direction=self.direction, length=self.length)
        with model.asfile() as f:
            mujoco_env.MujocoEnv.__init__(self, f.name, 5)

    def step(self, a):
        vec_dist = self.get_body_com("particle") - self.get_body_com("target")

        reward_dist = -np.linalg.norm(vec_dist)  # particle to target
        reward_ctrl = -np.square(a).sum()
        if self.no_reward:
            reward = 0
        elif self.sparse_reward:
            if reward_dist <= 0.1:
                reward = 1
            else:
                reward = 0
        else:
            reward = reward_dist + 0.001 * reward_ctrl

        self.do_simulation(a, self.frame_skip)
        ob = self._get_obs()
        self.episode_length += 1
        done = self.episode_length >= self.max_episode_length
        return ob, reward, done, dict(reward_dist=reward_dist, reward_ctrl=reward_ctrl)

    def viewer_setup(self):
        self.viewer.cam.trackbodyid = -1
        self.viewer.cam.distance = 4.0

    def reset_model(self):
        qpos = self.init_qpos
        qvel = self.init_qvel + self.np_random.uniform(
            size=self.model.nv, low=-0.01, high=0.01
        )
        self.set_state(qpos, qvel)
        self.episode_length = 0
        return self._get_obs()

    def _get_obs(self):
        obs = [self.get_body_com("particle")]
        if self.include_vel:
            obs.append(self.sim.data.get_body_xvelp("particle"))
        return np.concatenate(obs)

    def plot_trajs(self, *args, **kwargs):
        pass



"""Types and helper methods for transitions and trajectories."""

import dataclasses
import os
import pickle
from typing import Optional, Sequence

import numpy as np
import tensorflow as tf

from imitation.data import old_types


@dataclasses.dataclass(frozen=True)
class Trajectory:
    """A trajectory, e.g. a one episode rollout from an expert policy."""

    obs: np.ndarray
    """Observations, shape (trajectory_len + 1, ) + observation_shape."""

    acts: np.ndarray
    """Actions, shape (trajectory_len, ) + action_shape."""

    infos: Optional[np.ndarray]
    """An array of info dicts, length trajectory_len."""

    def __len__(self):
        """Returns number of transitions, `trajectory_len` in attribute docstrings.

        This is equal to the number of actions, and is always positive.
        """
        return len(self.acts)

    def __post_init__(self):
        """Performs input validation: check shapes are as specified in docstring."""
        if len(self.obs) != len(self.acts) + 1:
            raise ValueError(
                "expected one more observations than actions: "
                f"{len(self.obs)} != {len(self.acts)} + 1"
            )
        if self.infos is not None and len(self.infos) != len(self.acts):
            raise ValueError(
                "infos when present must be present for each action: "
                f"{len(self.infos)} != {len(self.acts)}"
            )
        if len(self.acts) == 0:
            raise ValueError("Degenerate trajectory: must have at least one action.")


def _rews_validation(rews: np.ndarray, acts: np.ndarray):
    if rews.shape != (len(acts),):
        raise ValueError(
            "rewards must be 1D array, one entry for each action: "
            f"{rews.shape} != ({len(acts)},)"
        )
    if rews.dtype not in [np.float32, np.float64, np.float128]:
        raise ValueError("rewards dtype {self.rews.dtype} not a float")


@dataclasses.dataclass(frozen=True)
class TrajectoryWithRew(Trajectory):
    rews: np.ndarray
    """Reward, shape (trajectory_len, ). dtype float."""

    def __post_init__(self):
        """Performs input validation, including for rews."""
        super().__post_init__()
        _rews_validation(self.rews, self.acts)


@dataclasses.dataclass(frozen=True)
class TransitionsMinimal:
    """A batch of obs-act transitions.

    This class and its subclasses are usually instantiated via
    `imitation.data.rollout.flatten_trajectories`.
    """

    obs: np.ndarray
    """
    Previous observations. Shape: (batch_size, ) + observation_shape.

    The i'th observation `obs[i]` in this array is the observation seen
    by the agent when choosing action `acts[i]`. `obs[i]` is not required to
    be from the timestep preceding `obs[i+1]`.
    """

    acts: np.ndarray
    """Actions. Shape: (batch_size,) + action_shape."""

    infos: np.ndarray
    """Array of info dicts. Shape: (batch_size,)."""

    def __len__(self):
        """Returns number of transitions. Always positive."""
        return len(self.obs)

    def __post_init__(self):
        """Performs input validation: check shapes & dtypes match docstring.

        Also make array values read-only.
        """
        for val in vars(self).values():
            if isinstance(val, np.ndarray):
                val.setflags(write=False)

        if len(self.obs) != len(self.acts):
            raise ValueError(
                "obs and acts must have same number of timesteps: "
                f"{len(self.obs)} != {len(self.acts)}"
            )
        if len(self.obs) == 0:
            raise ValueError("Must have non-zero number of observations.")

        if self.infos is not None and len(self.infos) != len(self.obs):
            raise ValueError(
                "obs and infos must have same number of timesteps: "
                f"{len(self.obs)} != {len(self.infos)}"
            )


@dataclasses.dataclass(frozen=True)
class Transitions(TransitionsMinimal):
    """A batch of obs-act-obs-done transitions."""

    next_obs: np.ndarray
    """New observation. Shape: (batch_size, ) + observation_shape.

    The i'th observation `next_obs[i]` in this array is the observation
    after the agent has taken action `acts[i]`.

    Invariants:
        * `next_obs.dtype == obs.dtype`
        * `len(next_obs) == len(obs)`
    """

    dones: np.ndarray
    """
    Boolean array indicating episode termination. Shape: (batch_size, ).

    `done[i]` is true iff `next_obs[i]` the last observation of an episode.
    """

    def __post_init__(self):
        """Performs input validation: check shapes & dtypes match docstring."""
        super().__post_init__()
        if self.obs.shape != self.next_obs.shape:
            raise ValueError(
                "obs and next_obs must have same shape: "
                f"{self.obs.shape} != {self.next_obs.shape}"
            )
        if self.obs.dtype != self.next_obs.dtype:
            raise ValueError(
                "obs and next_obs must have the same dtype: "
                f"{self.obs.dtype} != {self.next_obs.dtype}"
            )
        if self.dones.shape != (len(self.acts),):
            raise ValueError(
                "dones must be 1D array, one entry for each timestep: "
                f"{self.dones.shape} != ({len(self.acts)},)"
            )
        if self.dones.dtype != np.bool:
            raise ValueError(f"dones must be boolean, not {self.dones.dtype}")


@dataclasses.dataclass(frozen=True)
class TransitionsWithRew(Transitions):
    """A batch of obs-act-obs-rew-done transitions."""

    rews: np.ndarray
    """
    Reward. Shape: (batch_size, ). dtype float.

    The reward `rew[i]` at the i'th timestep is received after the
    agent has taken action `acts[i]`.
    """

    def __post_init__(self):
        """Performs input validation, including for rews."""
        super().__post_init__()
        _rews_validation(self.rews, self.acts)


def load(path: str) -> Sequence[TrajectoryWithRew]:
    """Loads a sequence of trajectories saved by `save()` from `path`."""
    # TODO(adam): remove backwards compatibility logic eventually (2021?)
    import sys

    try:
        assert "imitation.util.rollout" not in sys.modules
        sys.modules["imitation.util.rollout"] = old_types
        with open(path, "rb") as f:
            trajectories = pickle.load(f)
    finally:
        del sys.modules["imitation.util.rollout"]

    if len(trajectories) > 0:
        if isinstance(trajectories[0], old_types.Trajectory):
            trajectories = [
                TrajectoryWithRew(**traj._asdict()) for traj in trajectories
            ]

    return trajectories


def save(path: str, trajectories: Sequence[TrajectoryWithRew]) -> None:
    """Save a sequence of Trajectories to disk.

    Args:
        path: Trajectories are saved to this path.
        trajectories: The trajectories to save.
    """
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path + ".tmp", "wb") as f:
        pickle.dump(trajectories, f)
    # Ensure atomic write
    os.replace(path + ".tmp", path)
    tf.logging.info("Dumped demonstrations to {}.".format(path))



"""DAgger (https://arxiv.org/pdf/1011.0686.pdf).

Interactively trains policy by collecting some demonstrations, doing BC, collecting more
demonstrations, doing BC again, etc. Initially the demonstrations just come from the
expert's policy; over time, they shift to be drawn more and more from the imitator's
policy.
"""

import dataclasses
import os
from typing import Callable, Tuple

import cloudpickle
import gym
import numpy as np
import tensorflow as tf
from stable_baselines.common.policies import BasePolicy

from imitation.algorithms.bc import BC, set_tf_vars
from imitation.data import rollout, types
from imitation.util import util


def linear_beta_schedule(rampdown_rounds: int) -> Callable[[int], float]:
    """Linearly-decreasing schedule for beta (% of time that user act is used).

    Args:
      rampdown_rounds: number of rounds over which to anneal beta. Rounds
        are assumed to be numbered 0, 1, 2, ….

    Returns:
      schedule: function that takes current round number (0, or 1, or 2, etc.)
        as input & returns current beta as float.
    """
    assert rampdown_rounds > 0

    def schedule(i: int) -> float:
        assert i >= 0
        return min(1, max(0, (rampdown_rounds - i) / rampdown_rounds))

    return schedule


def _save_trajectory(
    npz_path: str,
    trajectory: types.Trajectory,
) -> None:
    """Save a trajectory as a compressed Numpy file."""
    save_dir = os.path.dirname(npz_path)
    if save_dir:
        os.makedirs(save_dir, exist_ok=True)
    assert isinstance(trajectory, types.Trajectory)
    np.savez_compressed(npz_path, **dataclasses.asdict(trajectory))


def _load_trajectory(npz_path: str) -> types.Trajectory:
    """Load a single trajectory from a compressed Numpy file."""
    np_data = np.load(npz_path, allow_pickle=True)
    has_rew = "rews" in np_data
    cls = types.TrajectoryWithRew if has_rew else types.Trajectory
    return cls(**dict(np_data.items()))


class InteractiveTrajectoryCollector(gym.Wrapper):
    """Wrapper around the `.step()` and `.reset()` of an env that allows DAgger to
    inject a "robot" action (i.e. an action from of the imitation policy) that overrides
    the action given to `.step()` when necessary.

    Will also automatically save trajectories.
    """

    def __init__(
        self,
        env: gym.Env,
        get_robot_act: Callable[[np.ndarray], np.ndarray],
        beta: float,
        save_dir: str,
    ):
        """Trajectory collector constructor.

        Args:
          env: environment to sample trajectories from.
          get_robot_act: get a single robot action that can be substituted for
              human action. Takes a single observation as input & returns a
              single action.
          beta: fraction of the time to use action given to .step() instead of
              robot action.
          save_dir: directory to save collected trajectories in.
        """
        super().__init__(env)
        self.get_robot_act = get_robot_act
        assert 0 <= beta <= 1
        self.beta = beta
        self.traj_accum = None
        self.save_dir = save_dir
        self._last_obs = None
        self._done_before = True
        self._is_reset = False

    def reset(self) -> np.ndarray:
        """Resets the environment.

        Returns:
            obs: first observation of a new trajectory.
        """
        self.traj_accum = rollout.TrajectoryAccumulator()
        obs = self.env.reset()
        self._last_obs = obs
        self.traj_accum.add_step({"obs": obs})
        self._done_before = False
        self._is_reset = True
        return obs

    def step(self, user_action: np.ndarray) -> Tuple[np.ndarray, float, bool, dict]:
        """Steps the environment.

        DAgger needs to be able to inject imitation policy actions randomly at some
        subset of time steps. This method will replace the given action with a
        "robot" (i.e. imitation policy) action if necessary.

        Args:
          user_action: the _intended_ demonstrator action for the current
            state. This will be executed with probability `self.beta`.
            Otherwise, a "robot" action will be sampled and executed
            instead.

        Returns:
          next_obs, reward, done, info: unchanged output of `self.env.step()`.
        """
        assert self._is_reset, "call .reset() before .step()"

        # Replace the given action with a robot action 100*(1-beta)% of the time.
        if np.random.uniform(0, 1) > self.beta:
            actual_act = self.get_robot_act(self._last_obs)
        else:
            actual_act = user_action

        # actually step the env & record data as appropriate
        next_obs, reward, done, info = self.env.step(actual_act)
        self._last_obs = next_obs
        self.traj_accum.add_step(
            {"acts": user_action, "obs": next_obs, "rews": reward, "infos": info}
        )

        # if we're finished, then save the trajectory & print a message
        if done and not self._done_before:
            trajectory = self.traj_accum.finish_trajectory()
            timestamp = util.make_unique_timestamp()
            trajectory_path = os.path.join(
                self.save_dir, "dagger-demo-" + timestamp + ".npz"
            )
            tf.logging.info(f"Saving demo at '{trajectory_path}'")
            _save_trajectory(trajectory_path, trajectory)

        if done:
            # record the fact that we're already done to avoid saving demo over and
            # over until the user resets
            self._done_before = True

        return next_obs, reward, done, info


class NeedsDemosException(Exception):
    """Signals demos need to be collected for current round before continuing."""


class DAggerTrainer:
    """Helper class for interactively training with DAgger.

    In essence, this is just BC with some helpers for incrementally
    resuming training and interpolating between demonstrator/learnt policies.
    Interaction proceeds in "rounds" in which the demonstrator first provides a
    fresh set of demonstrations, and then an underlying `BC` is invoked to
    fine-tune the policy on the entire set of demonstrations collected in all
    rounds so far. Demonstrations and policy/trainer checkpoints are stored in a
    directory with the following structure::

       scratch-dir-name/
           checkpoint-001.pkl
           checkpoint-002.pkl
           …
           checkpoint-XYZ.pkl
           checkpoint-latest.pkl
           demos/
               round-000/
                   demos_round_000_000.npz
                   demos_round_000_001.npz
                   …
               round-001/
                   demos_round_001_000.npz
                   …
               …
               round-XYZ/
                   …
    """

    SAVE_ATTRS = ("round_num",)
    DEMO_SUFFIX = ".npz"

    def __init__(
        self,
        env: gym.Env,
        scratch_dir: str,
        beta_schedule: Callable[[int], float] = None,
        **bc_kwargs,
    ):
        """Trainer constructor.

        Args:
          env: environment to train in.
          scratch_dir: directory to use to store intermediate training
              information (e.g. for resuming training).
          beta_schedule: provides a value of `beta` (the probability of taking
              expert action in any given state) at each round of training. If
              `None`, then `linear_beta_schedule` will be used instead.
          **bc_kwargs: additional arguments for constructing the `BC` that
              will be used to train the underlying policy.
        """
        # for pickling
        self._init_args = locals()
        self._init_args.update(bc_kwargs)
        del self._init_args["self"]
        del self._init_args["bc_kwargs"]

        if beta_schedule is None:
            beta_schedule = linear_beta_schedule(15)
        self.beta_schedule = beta_schedule
        self.scratch_dir = scratch_dir
        self.env = env
        self.round_num = 0
        self.bc_kwargs = bc_kwargs
        self._last_loaded_round = -1
        self._all_demos = []

        self._build_graph()

    def _build_graph(self):
        with tf.variable_scope("dagger"):
            self.bc_trainer = BC(
                self.env.observation_space, self.env.action_space, **self.bc_kwargs
            )
            with self._graph.as_default():
                self._vars = tf.get_collection(
                    tf.GraphKeys.GLOBAL_VARIABLES, scope=tf.get_variable_scope().name
                )
            assert len(self._vars) > 0

    def _load_all_demos(self):
        num_demos_by_round = []
        for round_num in range(self._last_loaded_round + 1, self.round_num + 1):
            round_dir = self._demo_dir_path_for_round(round_num)
            demo_paths = self._get_demo_paths(round_dir)
            self._all_demos.extend(_load_trajectory(p) for p in demo_paths)
            num_demos_by_round.append(len(demo_paths))
        tf.logging.info(f"Loaded {len(self._all_demos)} total")
        demo_transitions = rollout.flatten_trajectories(self._all_demos)
        return demo_transitions, num_demos_by_round

    def _get_demo_paths(self, round_dir):
        return [
            os.path.join(round_dir, p)
            for p in os.listdir(round_dir)
            if p.endswith(self.DEMO_SUFFIX)
        ]

    def _demo_dir_path_for_round(self, round_num=None):
        if round_num is None:
            round_num = self.round_num
        return os.path.join(self.scratch_dir, "demos", f"round-{round_num:03d}")

    def _check_has_latest_demos(self):
        demo_dir = self._demo_dir_path_for_round()
        demo_paths = self._get_demo_paths(demo_dir) if os.path.isdir(demo_dir) else []
        if len(demo_paths) == 0:
            raise NeedsDemosException(
                f"No demos found for round {self.round_num} in dir '{demo_dir}'. "
                f"Maybe you need to collect some demos? See "
                f".get_trajectory_collector()"
            )

    def _try_load_demos(self):
        self._check_has_latest_demos()
        if self._last_loaded_round < self.round_num:
            transitions, num_demos = self._load_all_demos()
            tf.logging.info(
                f"Loaded {sum(num_demos)} new demos from {len(num_demos)} rounds"
            )
            self.bc_trainer.set_expert_dataset(transitions)
            self._last_loaded_round = self.round_num

    @property
    def _sess(self):
        return self.bc_trainer.sess

    @property
    def _graph(self):
        return self.bc_trainer.sess.graph

    def extend_and_update(self, **train_kwargs) -> int:
        """Extend internal batch of data and train.

        Specifically, this method will load new transitions (if necessary), train
        the model for a while, and advance the round counter. If there are no fresh
        demonstrations in the demonstration directory for the current round, then
        this will raise a `NeedsDemosException` instead of training or advancing
        the round counter. In that case, the user should call
        `.get_trajectory_collector()` and use the returned
        `InteractiveTrajectoryCollector` to produce a new set of demonstrations for
        the current interaction round.

        Arguments:
          **train_kwargs: arguments to pass to `BC.train()`.

        Returns:
          round_num: new round number after advancing the round counter.
        """
        tf.logging.info("Loading demonstrations")
        self._try_load_demos()
        tf.logging.info(f"Training at round {self.round_num}")
        self.bc_trainer.train(**train_kwargs)
        self.round_num += 1
        tf.logging.info(f"New round number is {self.round_num}")
        return self.round_num

    def get_trajectory_collector(self) -> InteractiveTrajectoryCollector:
        """Create trajectory collector to extend current round's demonstration set.

        Returns:
          collector: an `InteractiveTrajectoryCollector` configured with the
            appropriate beta, appropriate imitator policy, etc. for the current
            round. Refer to the documentation for
            `InteractiveTrajectoryCollector` to see how to use this.
        """
        save_dir = self._demo_dir_path_for_round()
        beta = self.beta_schedule(self.round_num)

        def get_robot_act(obs):
            (act,), _, _, _ = self.bc_trainer.policy.step(obs[None])
            return act

        collector = InteractiveTrajectoryCollector(
            env=self.env, get_robot_act=get_robot_act, beta=beta, save_dir=save_dir
        )

        return collector

    def save_trainer(self) -> str:
        """Create a snapshot of trainer in the scratch/working directory.

        The created snapshot can also be reloaded with `.reconstruct_trainer()`.
        For convenience, this method will also create a separate snapshot of the
        current policy, which can then be passed to evaluation routines for other
        algorithms.

        Returns:
          checkpoint_path: a path to one of the created `DAggerTrainer`
            checkpoints.
        """
        os.makedirs(self.scratch_dir, exist_ok=True)
        # save full trainer checkpoints
        checkpoint_paths = [
            os.path.join(self.scratch_dir, f"checkpoint-{self.round_num:03d}.pkl"),
            os.path.join(self.scratch_dir, "checkpoint-latest.pkl"),
        ]
        saved_attrs = {
            attr_name: getattr(self, attr_name) for attr_name in self.SAVE_ATTRS
        }
        snapshot_dict = {
            "init_args": self._init_args,
            "variable_values": self._sess.run(self._vars),
            "saved_attrs": saved_attrs,
        }
        for checkpoint_path in checkpoint_paths:
            with open(checkpoint_path, "wb") as fp:
                cloudpickle.dump(snapshot_dict, fp)

        # save policies separately for convenience
        policy_paths = [
            os.path.join(self.scratch_dir, f"policy-{self.round_num:03d}.pkl"),
            os.path.join(self.scratch_dir, "policy-latest.pkl"),
        ]
        for policy_path in policy_paths:
            self.save_policy(policy_path)

        return checkpoint_paths[-1]

    @classmethod
    def reconstruct_trainer(cls, scratch_dir: str) -> "DAggerTrainer":
        """Reconstruct trainer from the latest snapshot in some working directory.

        Args:
          scratch_dir: path to the working directory created by a previous run of
            this algorithm. The directory should contain a
            `checkpoint-latest.pkl` file.

        Returns:
          trainer: a reconstructed `DAggerTrainer` with the same state as the
            previously-saved one.
        """
        checkpoint_path = os.path.join(scratch_dir, "checkpoint-latest.pkl")
        with open(checkpoint_path, "rb") as fp:
            saved_trainer = cloudpickle.load(fp)
        # reconstruct from old init args
        trainer = cls(**saved_trainer["init_args"])
        # set TF variables
        set_tf_vars(
            values=saved_trainer["variable_values"],
            tf_vars=trainer._vars,
            sess=trainer.bc_trainer.sess,
        )
        for attr_name in cls.SAVE_ATTRS:
            attr_value = saved_trainer["saved_attrs"][attr_name]
            setattr(trainer, attr_name, attr_value)
        return trainer

    def save_policy(self, *args, **kwargs):
        """Save the current policy only, (and not the rest of the trainer).

        Refer to docs for `BC.save_policy`.
        """
        self.bc_trainer.save_policy(*args, **kwargs)

    @staticmethod
    def reconstruct_policy(*args, **kwargs) -> BasePolicy:
        """Reconstruct a policy saved with `save_policy()`.

        This is an alias for `BC.reconstruct_policy()`.
        """
        return BC.reconstruct_policy(*args, **kwargs)



import contextlib
import os
from typing import ContextManager, Optional, Sequence

import stable_baselines.logger as sb_logger


def _build_output_formats(
    folder: str,
    format_strs: Sequence[str] = None,
) -> Sequence[sb_logger.KVWriter]:
    """Build output formats for initializing a Stable Baselines Logger.

    Args:
      folder: Path to directory that logs are written to.
      format_strs: An list of output format strings. For details on available
        output formats see `stable_baselines.logger.make_output_format`.
    """
    os.makedirs(folder, exist_ok=True)
    output_formats = [sb_logger.make_output_format(f, folder) for f in format_strs]
    return output_formats


class _HierarchicalLogger(sb_logger.Logger):
    def __init__(
        self,
        default_logger: sb_logger.Logger,
        format_strs: Sequence[str] = ("stdout", "log", "csv"),
    ):
        """A logger with a context for accumulating mean values.

        Args:
          default_logger: The default logger when not in the a `accumulate_means`
            context. Also the logger to which mean values are written to when
            contexts are over.
          format_strs: An list of output format strings that should be used by
            every Logger initialized by this class during an `AccumulatingMeans`
            context. For details on available output formats see
            `stable_baselines.logger.make_output_format`.
        """
        self.default_logger = default_logger
        self.current_logger = None
        self._cached_loggers = {}
        self._subdir = None
        self.format_strs = format_strs
        super().__init__(folder=self.default_logger.dir, output_formats=None)

    @contextlib.contextmanager
    def accumulate_means(self, subdir: str):
        """Temporarily modifies this _HierarchicalLogger to accumulate means values.

        During this context, `self.logkv(key, value)` writes the "raw" values in
        "{self.default_logger.log_dir}/{subdir}" under the key "raw/{subdir}/{key}".
        At the same time, any call to `self.logkv` will also accumulate mean values
        on the default logger by calling
        `self.default_logger.logkv_mean(f"mean/{subdir}/{key}", value)`.

        During the context, `self.logkv(key, value)` will write the "raw" values in
        `"{self.default_logger.log_dir}/subdir"` under the key "raw/{subdir}/key".

        After the context exits, calling `self.dumpkvs()` will write the means
        of all the "raw" values accumulated during this context to
        `self.default_logger` under keys with the prefix `mean/{subdir}/`

        Note that the behavior of other logging methods, `log` and `logkv_mean`
        are unmodified and will go straight to the default logger.

        Args:
          subdir: A string key which determines the `folder` where raw data is
            written and temporary logging prefixes for raw and mean data. Entering
            an `accumulate_means` context in the future with the same `subdir`
            will safely append to logs written in this folder rather than
            overwrite.
        """
        if self.current_logger is not None:
            raise RuntimeError("Nested `accumulate_means` context")

        if subdir in self._cached_loggers:
            logger = self._cached_loggers[subdir]
        else:
            folder = os.path.join(self.default_logger.dir, "raw", subdir)
            os.makedirs(folder)
            output_formats = _build_output_formats(folder, self.format_strs)
            logger = sb_logger.Logger(folder, output_formats)
            self._cached_loggers[subdir] = logger

        try:
            self.current_logger = logger
            self._subdir = subdir
            yield
        finally:
            self.current_logger = None
            self._subdir = None

    def logkv(self, key, val):
        if self.current_logger is not None:
            assert self._subdir is not None
            raw_key = os.path.join("raw", self._subdir, key)
            self.current_logger.logkv(raw_key, val)

            mean_key = os.path.join("mean", self._subdir, key)
            self.default_logger.logkv_mean(mean_key, val)
        else:
            self.default_logger.logkv_mean(key, val)

    @property
    def _logger(self):
        if self.current_logger is not None:
            return self.current_logger
        else:
            return self.default_logger

    def dumpkvs(self):
        self._logger.dumpkvs()

    def get_dir(self) -> str:
        return self._logger.get_dir()

    def log(self, *args, **kwargs):
        self.default_logger.log(*args, **kwargs)

    def logkv_mean(self, key, val):
        self.default_logger.logkv_mean(key, val)

    def close(self):
        raise NotImplementedError


def _sb_logger_configure_replacement(*args, **kwargs):
    raise RuntimeError(
        "Shouldn't call stable_baselines.logger.configure "
        "once imitation.logger.configure() has been called"
    )


def _sb_logger_reset_replacement():
    raise RuntimeError(
        "Shouldn't call stable_baselines.logger.reset "
        "once imitation.logger.configure() has been called"
    )


def is_configured() -> bool:
    """Return True if the custom logger is active."""
    return isinstance(sb_logger.Logger.CURRENT, _HierarchicalLogger)


def configure(folder: str, format_strs: Optional[Sequence[str]] = None) -> None:
    """Configure Stable Baselines logger to be `accumulate_means()`-compatible.

    After this function is called, `stable_baselines.logger.{configure,reset}()`
    are replaced with stubs that raise RuntimeError.

    Args:
        folder: Argument from `stable_baselines.logger.configure`.
        format_strs: An list of output format strings. For details on available
          output formats see `stable_baselines.logger.make_output_format`.
    """
    # Replace `stable_baselines.logger` methods with erroring stubs to
    # prevent unexpected logging state from mixed logging configuration.
    sb_logger.configure = _sb_logger_configure_replacement
    sb_logger.reset = _sb_logger_reset_replacement

    if format_strs is None:
        format_strs = ["stdout", "log", "csv"]
    output_formats = _build_output_formats(folder, format_strs)
    default_logger = sb_logger.Logger(folder, output_formats)
    hier_logger = _HierarchicalLogger(default_logger, format_strs)
    sb_logger.Logger.CURRENT = hier_logger
    sb_logger.log("Logging to %s" % folder)
    assert is_configured()


def logkv(key, val) -> None:
    """Alias for `stable_baselines.logger.logkv`."""
    sb_logger.logkv(key, val)


def dumpkvs() -> None:
    """Alias for `stable_baselines.logger.dumpkvs`."""
    sb_logger.dumpkvs()


def accumulate_means(subdir_name: str) -> ContextManager:
    """Temporarily redirect logkv() to a different logger and auto-track kvmeans.

    Within this context, the original logger is swapped out for a special logger
    in directory `"{current_logging_dir}/raw/{subdir_name}"`.

    The special logger's `stable_baselines.logger.logkv(key, val)`, in addition
    to tracking its own logs, also forwards the log to the original logger's
    `.logkv_mean()` under the key `mean/{subdir_name}/{key}`.

    After the context exits, these means can be dumped as usual using
    `stable_baselines.logger.dumpkvs()` or `imitation.util.logger.dumpkvs()`.

    Note that the behavior of other logging methods, `log` and `logkv_mean`
    are unmodified and will go straight to the original logger.

    This context cannot be nested.

    Args:
      subdir_name: A string key for building the logger, as described above.

    Returns:
      A context manager.
    """
    assert is_configured()
    hier_logger = sb_logger.Logger.CURRENT  # type: _HierarchicalLogger
    return hier_logger.accumulate_means(subdir_name)



"""Custom policy classes and convenience methods."""

import abc

import gym
import numpy as np
from stable_baselines.common import BaseRLModel
from stable_baselines.common.policies import BasePolicy, FeedForwardPolicy


class HardCodedPolicy(BasePolicy, abc.ABC):
    """Abstract class for hard-coded (non-trainable) policies."""

    def __init__(self, ob_space: gym.Space, ac_space: gym.Space):
        self.ob_space = ob_space
        self.ac_space = ac_space
        self.n_env = None
        self.n_steps = None
        self.n_batch = None

    def step(self, obs, state=None, mask=None, deterministic=False):
        actions = []
        for ob in obs:
            assert self.ob_space.contains(ob)
            actions.append(self._choose_action(obs))
        return np.array(actions), None, None, None

    @abc.abstractmethod
    def _choose_action(self, obs: np.ndarray) -> np.ndarray:
        """Chooses an action, optionally based on observation obs."""

    def proba_step(self, obs, state=None, mask=None):
        raise NotImplementedError()


class RandomPolicy(HardCodedPolicy):
    """Returns random actions."""

    def __init__(self, ob_space: gym.Space, ac_space: gym.Space):
        super().__init__(ob_space, ac_space)

    def _choose_action(self, obs: np.ndarray) -> np.ndarray:
        return self.ac_space.sample()


class ZeroPolicy(HardCodedPolicy):
    """Returns constant zero action."""

    def __init__(self, ob_space: gym.Space, ac_space: gym.Space):
        super().__init__(ob_space, ac_space)

    def _choose_action(self, obs: np.ndarray) -> np.ndarray:
        return np.zeros(self.ac_space.shape, dtype=self.ac_space.dtype)


class FeedForward32Policy(FeedForwardPolicy):
    """A feed forward policy network with two hidden layers of 32 units.

    This matches the IRL policies in the original AIRL paper.

    Note: This differs from stable_baselines MlpPolicy in two ways: by having
    32 rather than 64 units, and by having policy and value networks share weights
    except at the final layer.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, net_arch=[32, 32], feature_extraction="mlp")


def get_action_policy(policy, observation, deterministic=True):
    """Gets an action from a Stable Baselines policy after some processing.

    Specifically, clips actions to the action space associated with `policy` and
    automatically accounts for vectorized environments inputs.

    This code was adapted from Stable Baselines' `BaseRLModel.predict()`.

    Args:
        policy (stable_baselines.common.policies.BasePolicy): The policy.
        observation (np.ndarray): The input to the policy network. Can either
            be a single input with shape `policy.ob_space.shape` or a vectorized
            input with shape `(n_batch,) + policy.ob_space.shape`.
        deterministic (bool): Whether or not to return deterministic actions
            (usually means argmax over policy's action distribution).

    Returns:
       action (np.ndarray): The action output of the policy network. If
           `observation` is not vectorized (has shape `policy.ob_space.shape`
           instead of shape `(n_batch,) + policy.ob_space.shape`) then
           `action` has shape `policy.ac_space.shape`.
           Otherwise, `action` has shape `(n_batch,) + policy.ac_space.shape`.
    """
    observation = np.array(observation)
    vectorized_env = BaseRLModel._is_vectorized_observation(
        observation, policy.ob_space
    )

    observation = observation.reshape((-1,) + policy.ob_space.shape)
    actions, _, states, _ = policy.step(observation, deterministic=deterministic)

    clipped_actions = actions
    if isinstance(policy.ac_space, gym.spaces.Box):
        clipped_actions = np.clip(actions, policy.ac_space.low, policy.ac_space.high)

    if not vectorized_env:
        clipped_actions = clipped_actions[0]

    return clipped_actions, states



"""Environments used for testing and benchmarking.

These are not a core part of the imitation package. They are relatively lightly tested,
and may be changed without warning.
"""

# Register environments with Gym
from imitation.envs.examples import airl_envs, model_envs  # noqa: F401



"""Train GAIL or AIRL.

Can be used as a CLI script, or the `train_and_plot` function can be called directly.
"""

import os
import os.path as osp
from typing import Mapping, Optional

import tensorflow as tf
from sacred.observers import FileStorageObserver

from imitation.algorithms import adversarial
from imitation.data import rollout, types
from imitation.policies import serialize
from imitation.scripts.config.train_adversarial import train_ex
from imitation.util import logger, networks
from imitation.util import sacred as sacred_util
from imitation.util import util


def save(trainer, save_path):
    """Save discriminator and generator."""
    # We implement this here and not in Trainer since we do not want to actually
    # serialize the whole Trainer (including e.g. expert demonstrations).
    trainer.discrim.save(os.path.join(save_path, "discrim"))
    # TODO(gleave): unify this with the saving logic in data_collect?
    # (Needs #43 to be merged before attempting.)
    serialize.save_stable_model(
        os.path.join(save_path, "gen_policy"),
        trainer.gen_policy,
        trainer.venv_train_norm,
    )


@train_ex.main
def train(
    _run,
    _seed: int,
    algorithm: str,
    env_name: str,
    num_vec: int,
    parallel: bool,
    max_episode_steps: Optional[int],
    rollout_path: str,
    n_expert_demos: Optional[int],
    log_dir: str,
    total_timesteps: int,
    n_episodes_eval: int,
    init_tensorboard: bool,
    checkpoint_interval: int,
    init_rl_kwargs: Mapping,
    algorithm_kwargs: Mapping[str, Mapping],
    discrim_net_kwargs: Mapping[str, Mapping],
) -> dict:
    """Train an adversarial-network-based imitation learning algorithm.

    Checkpoints:
        - DiscrimNets are saved to `f"{log_dir}/checkpoints/{step}/discrim/"`,
            where step is either the training epoch or "final".
        - Generator policies are saved to `f"{log_dir}/checkpoints/{step}/gen_policy/"`.

    Args:
        _seed: Random seed.
        algorithm: A case-insensitive string determining which adversarial imitation
            learning algorithm is executed. Either "airl" or "gail".
        env_name: The environment to train in.
        num_vec: Number of `gym.Env` to vectorize.
        parallel: Whether to use "true" parallelism. If True, then use `SubProcVecEnv`.
            Otherwise, use `DummyVecEnv` which steps through environments serially.
        max_episode_steps: If not None, then a TimeLimit wrapper is applied to each
            environment to artificially limit the maximum number of timesteps in an
            episode.
        rollout_path: Path to pickle containing list of Trajectories. Used as
            expert demonstrations.
        n_expert_demos: The number of expert trajectories to actually use
            after loading them from `rollout_path`.
            If None, then use all available trajectories.
            If `n_expert_demos` is an `int`, then use exactly `n_expert_demos`
            trajectories, erroring if there aren't enough trajectories. If there are
            surplus trajectories, then use the first `n_expert_demos` trajectories and
            drop the rest.
        log_dir: Directory to save models and other logging to.
        total_timesteps: The number of transitions to sample from the environment
            during training.
        n_episodes_eval: The number of episodes to average over when calculating
            the average episode reward of the imitation policy for return.
        init_tensorboard: If True, then write tensorboard logs to `{log_dir}/sb_tb`.
        checkpoint_interval: Save the discriminator and generator models every
            `checkpoint_interval` epochs and after training is complete. If 0,
            then only save weights after training is complete. If <0, then don't
            save weights at all.
        init_rl_kwargs: Keyword arguments for `init_rl`, the RL algorithm initialization
            utility function.
        algorithm_kwargs: Keyword arguments for the `GAIL` or `AIRL` constructor
            that can apply to either constructor. Unlike a regular kwargs argument, this
            argument can only have the following keys: "shared", "airl", and "gail".

            `algorithm_kwargs["airl"]`, if it is provided, is a kwargs `Mapping` passed
            to the `AIRL` constructor when `algorithm == "airl"`. Likewise
            `algorithm_kwargs["gail"]` is passed to the `GAIL` constructor when
            `algorithm == "gail"`. `algorithm_kwargs["shared"]`, if provided, is passed
            to both the `AIRL` and `GAIL` constructors. Duplicate keyword argument keys
            between `algorithm_kwargs["shared"]` and `algorithm_kwargs["airl"]` (or
            "gail") leads to an error.
        discrim_net_kwargs: Keyword arguments for the `DiscrimNet` constructor. Unlike a
            regular kwargs argument, this argument can only have the following keys:
            "shared", "airl", "gail". These keys have the same meaning as they do in
            `algorithm_kwargs`.

    Returns:
        A dictionary with two keys. "imit_stats" gives the return value of
        `rollout_stats()` on rollouts test-reward-wrapped environment, using the final
        policy (remember that the ground-truth reward can be recovered from the
        "monitor_return" key). "expert_stats" gives the return value of
        `rollout_stats()` on the expert demonstrations loaded from `rollout_path`.
    """
    assert os.path.exists(rollout_path)
    total_timesteps = int(total_timesteps)

    tf.logging.info("Logging to %s", log_dir)
    logger.configure(log_dir, ["tensorboard", "stdout"])
    os.makedirs(log_dir, exist_ok=True)
    sacred_util.build_sacred_symlink(log_dir, _run)

    expert_trajs = types.load(rollout_path)
    if n_expert_demos is not None:
        assert len(expert_trajs) >= n_expert_demos
        expert_trajs = expert_trajs[:n_expert_demos]
    expert_transitions = rollout.flatten_trajectories(expert_trajs)

    with networks.make_session():
        if init_tensorboard:
            tensorboard_log = osp.join(log_dir, "sb_tb")
        else:
            tensorboard_log = None

        venv = util.make_vec_env(
            env_name,
            num_vec,
            seed=_seed,
            parallel=parallel,
            log_dir=log_dir,
            max_episode_steps=max_episode_steps,
        )

        # TODO(shwang): Let's get rid of init_rl later on?
        # It's really just a stub function now.
        gen_policy = util.init_rl(
            venv, verbose=1, tensorboard_log=tensorboard_log, **init_rl_kwargs
        )

        # Convert Sacred's ReadOnlyDict to dict so we can modify it.
        allowed_keys = {"shared", "gail", "airl"}
        assert discrim_net_kwargs.keys() <= allowed_keys
        assert algorithm_kwargs.keys() <= allowed_keys

        discrim_kwargs_shared = discrim_net_kwargs.get("shared", {})
        discrim_kwargs_algo = discrim_net_kwargs.get(algorithm, {})
        final_discrim_kwargs = dict(**discrim_kwargs_shared, **discrim_kwargs_algo)

        algorithm_kwargs_shared = algorithm_kwargs.get("shared", {})
        algorithm_kwargs_algo = algorithm_kwargs.get(algorithm, {})
        final_algorithm_kwargs = dict(
            **algorithm_kwargs_shared,
            **algorithm_kwargs_algo,
        )

        if algorithm.lower() == "gail":
            algo_cls = adversarial.GAIL
        elif algorithm.lower() == "airl":
            algo_cls = adversarial.AIRL
        else:
            raise ValueError(f"Invalid value algorithm={algorithm}.")

        trainer = algo_cls(
            venv=venv,
            expert_data=expert_transitions,
            gen_policy=gen_policy,
            log_dir=log_dir,
            discrim_kwargs=final_discrim_kwargs,
            **final_algorithm_kwargs,
        )

        def callback(epoch):
            if checkpoint_interval > 0 and epoch % checkpoint_interval == 0:
                save(trainer, os.path.join(log_dir, "checkpoints", f"{epoch:05d}"))

        trainer.train(total_timesteps, callback)

        # Save final artifacts.
        if checkpoint_interval >= 0:
            save(trainer, os.path.join(log_dir, "checkpoints", "final"))

        # Final evaluation of imitation policy.
        results = {}
        sample_until_eval = rollout.min_episodes(n_episodes_eval)
        trajs = rollout.generate_trajectories(
            trainer.gen_policy, trainer.venv_test, sample_until=sample_until_eval
        )
        results["expert_stats"] = rollout.rollout_stats(expert_trajs)
        results["imit_stats"] = rollout.rollout_stats(trajs)
        return results


def main_console():
    observer = FileStorageObserver(osp.join("output", "sacred", "train"))
    train_ex.observers.append(observer)
    train_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



from typing import NamedTuple, Tuple

import numpy as np
from gym.envs.mujoco import mujoco_env
from gym.spaces import Box


class MapConfig(NamedTuple):
    xs: Tuple[float, float] = (-0.3, 0.3)
    ys: Tuple[float, float] = (-0.3, 0.3)
    xres: int = 50
    yres: int = 50


def make_heat_map(eval_func, map_config):
    gps = get_dense_gridpoints(map_config)
    vals = np.zeros(map_config.xres * map_config.yres)
    for i, pnt in enumerate(gps):
        vals[i] = eval_func(pnt)
    return predictions_to_heatmap(vals, map_config)


def get_dense_gridpoints(map_config):
    xl = np.linspace(map_config.xs[0], map_config.xs[1], num=map_config.xres)
    yl = np.linspace(map_config.ys[0], map_config.ys[1], num=map_config.yres)
    gridpoints = np.zeros((map_config.xres * map_config.yres, 2))
    for i in range(map_config.xres):
        for j in range(map_config.yres):
            gridpoints[i + map_config.xres * j] = np.array((xl[i], yl[j]))
    return gridpoints


def predictions_to_heatmap(predictions, map_config):
    res = np.zeros((map_config.xres, map_config.yres))
    for i in range(map_config.xres):
        for j in range(map_config.yres):
            res[i, j] = predictions[i + map_config.xres * j]
    res = res / np.max(res)
    return res.T


def make_density_map(paths, map_config):
    xs = np.linspace(map_config.xs[0], map_config.xs[1], num=map_config.xres + 1)
    ys = np.linspace(map_config.ys[0], map_config.ys[1], num=map_config.yres + 1)
    y = paths[:, 0]
    x = paths[:, 1]
    H, xedges, yedges = np.histogram2d(y, x, bins=(xs, ys))
    H = H.astype(np.float)
    H = H / np.max(H)
    return H.T


def plot_maps(combined_list=None, *heatmaps):
    import matplotlib.pyplot as plt

    combined = np.c_[heatmaps]
    if combined_list is not None:
        combined_list.append(combined)
        combined = np.concatenate(combined_list)
    else:
        combined_list = []
    plt.figure()
    plt.imshow(combined, cmap="afmhot", interpolation="none")
    plt.show()
    return combined_list


class TwoDEnv(mujoco_env.MujocoEnv):
    def __init__(self, model_path, frame_skip, xbounds, ybounds):
        super(TwoDEnv, self).__init__(model_path=model_path, frame_skip=frame_skip)
        assert isinstance(self.observation_space, Box)
        assert self.observation_space.shape == (2,)
        self.__map_config = MapConfig(
            xs=(xbounds[0], xbounds[1]), ys=(ybounds[0], ybounds[1])
        )

    @property
    def map_config(self):
        return self.__map_config

    @property
    def grid_flat_dim(self):
        return self.map_config.xres * self.map_config.yres

    def make_density_map(self, paths):
        return make_density_map(paths, self.map_config)

    def make_heatmap(self, eval_func):
        return make_heat_map(eval_func, self.map_config)

    def get_dense_gridpoints(self):
        return get_dense_gridpoints(self.map_config)

    def predictions_to_heatmap(self, predictions):
        return predictions_to_heatmap(predictions, self.map_config)
